// RANSAC.cpp : This file contains the 'main' function. Program execution begins and ends there.
//

#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>

#include <iostream>
#include <random>
#include <ctime>
#include <vector>

int WIDTH = 640;
int HEIGHT = 480;

std::mt19937 rng;
std::uniform_real_distribution<float> udist(0.f, 1.f);
std::normal_distribution<float> ndist(0.f, 5.f);

void init_random()
{
    rng.seed(static_cast<long unsigned int>(time(0)));
}

float random_number()
{
    return udist(rng);
}

float random_noise()
{
    return ndist(rng);
}

cv::Point2f randomPoint()
{
    float x = random_number() * WIDTH;
    float y = random_number() * HEIGHT;
    return { x,y };
}

cv::Point2f randomDir()
{
    float angle = random_number() * CV_2PI;
    return { cos(angle), sin(angle) };
}

cv::Point3f parametricToImplicit(const cv::Point2f& p, const cv::Point2f& dir)
{
    float a = -dir.y;
    float b = dir.x;
    float c = -(a * p.x + b * p.y);

    return cv::Point3f(a, b, c);
}

void newParametricAndT(const cv::Point2f & p, const cv::Point2f & dir, cv::Point2f& newP, cv::Point2f& newDir, float& Tmax)
{
    cv::Point3f l = parametricToImplicit(p, dir);
    float a = l.x;
    float b = l.y;
    float c = l.z;

    cv::Point2f p1(0, -c / b);
    cv::Point2f p2(WIDTH, (-a*WIDTH - c)/b);
    cv::Point2f p3(-c/a, 0);
    cv::Point2f p4((-b*HEIGHT-c)/a, HEIGHT);
    std::vector<cv::Point2f> points = { p1, p2, p3, p4 };

    cv::Point2f P, Q;
    int i = 0;
    for (; i < 4; ++i)
    {
        if ((0 <= points[i].x && points[i].x <= WIDTH) && (0 <= points[i].y && points[i].y <= HEIGHT)) {
            P = points[i];
            break;
        }
    }
    ++i;
    for (; i < 4; ++i)
    {
        if ((0 <= points[i].x && points[i].x <= WIDTH) && (0 <= points[i].y && points[i].y <= HEIGHT)) {
            Q = points[i];
            break;
        }
    }

    cv::Point2f v = Q - P;
    Tmax = cv::norm(v);
    v /= cv::norm(v);

    newP = P;
    newDir = v;
}

void generateData(std::vector<cv::Point2f>& points, int N, cv::Point2f p, cv::Point2f dir, float inlier_ratio)
{
    int N_inlier = N * inlier_ratio;
    int N_outlier = N - N_inlier;

    float Tmax;
    newParametricAndT(p, dir, p, dir, Tmax); // We make sure to only generate line points inside the image


    for (int i = 0; i < N_inlier; ++i)
    {
        float t = random_number() * Tmax;
        cv::Point2f q = p + t * dir;
        q.x += random_noise();
        q.y += random_noise();

        points.push_back(q);
    }

    for (int i = 0; i < N_outlier; ++i)
    {
        cv::Point2f q;
        q.x = random_number() * WIDTH;
        q.y = random_number() * HEIGHT;
        points.push_back(q);
    }

}

cv::Point3f fitLine(std::vector<cv::Point2f>& points)
{
    cv::Mat X(points.size(), 3, CV_32F);
    for (int i = 0; i < points.size(); ++i)
    {
        X.at<float>(i, 0) = points[i].x;
        X.at<float>(i, 1) = points[i].y;
        X.at<float>(i, 2) = 1.f;
    }

    cv::Mat XtX = X.t() * X;

    cv::Mat evals, evecs;
    cv::eigen(XtX, evals, evecs);

    cv::Point3f l;
    l.x = evecs.at<float>(2, 0);
    l.y = evecs.at<float>(2, 1);
    l.z = evecs.at<float>(2, 2);

    // Normalization
    float n = sqrt(l.x * l.x + l.y * l.y);
    l /= n;

    return l;
}

cv::Point3f fitLine(cv::Point2f p1, cv::Point2f p2)
{
    cv::Point2f dir = p2 - p1;
    dir /= sqrt(dir.x * dir.x + dir.y * dir.y); // normalization

    cv::Point3f l;
    l.x = -dir.y;
    l.y = dir.x;
    l.z = -l.x * p1.x - l.y * p1.y;

    return l;
}

float distance(cv::Point3f& l, cv::Point2f p)
{
    return abs(l.x * p.x + l.y * p.y + l.z);
}

int iterationNumber(float confidence, int sampleSize, int inliers, int N)
{
    float a = std::log(1.0f - confidence);
    float ratio = (float)inliers / (float)N;
    float b = std::log(1.0f - std::pow(ratio, sampleSize));

    //if (b < std::numeric_limits<float>::epsilon()) return std::numeric_limits<int>::max();

    return std::ceil(a / b);
}

cv::Point3f RANSAC(std::vector<cv::Point2f>& points, int max_iterations, float sigma)
{
    int best_inliers_nr = 0;
    cv::Point3f best_model;
    std::vector<cv::Point2f> best_inliers;
    std::cout << "New max iters: " << max_iterations << std::endl;
    for (int iter = 0; iter < max_iterations; ++iter)
    {
        // 1. Select m points randomly
        int idx1 = random_number() * (points.size() - 1);
        int idx2 = random_number() * (points.size() - 1);
        while (idx1 == idx2)
        {
            idx2 = random_number() * (points.size() - 1);
        }

        // 2. Fit a model on the sample
        cv::Point3f current_model = fitLine(points[idx1], points[idx2]);

        // 3. Calculating the distance of the points
        int inliers_nr = 0;
        std::vector<cv::Point2f> current_inliers;
        for (int i = 0; i < points.size(); ++i)
        {
            if (distance(current_model, points[i]) < sigma)
            {
                current_inliers.push_back(points[i]);
                ++inliers_nr;
            }
        }

        // 4. Save the best model so far
        if (inliers_nr > best_inliers_nr)
        {
            best_inliers_nr = inliers_nr;
            best_model = current_model;
            best_inliers = current_inliers;

            max_iterations = std::min(max_iterations, iterationNumber(0.95, 2, best_inliers_nr, points.size()));
            //std::cout << "New max iters: " << max_iterations << std::endl;
        }
    }

    best_model = fitLine(best_inliers);
    return best_model;
}

std::vector<cv::Point3f> MultiRANSAC(std::vector<cv::Point2f>& points, int max_iterations, float sigma, int N)
{
    int best_inliers_nr = 0;
    std::vector<cv::Point3f> best_model;
    std::vector<std::vector<cv::Point2f>> best_inliers;
    for (int iter = 0; iter < max_iterations; ++iter)
    {
        std::vector<cv::Point3f> current_model(N);
        for (int i = 0; i < N; ++i) {
            // 1. Select m points randomly
            int idx1 = random_number() * (points.size() - 1);
            int idx2 = random_number() * (points.size() - 1);
            while (idx1 == idx2)
            {
                idx2 = random_number() * (points.size() - 1);
            }

            // 2. Fit a model on the sample
            current_model[i] = fitLine(points[idx1], points[idx2]);
        }
        // 3. Calculating the distance of the points
        int inliers_nr = 0;
        std::vector<std::vector<cv::Point2f>> current_inliers(N);
        for (int i = 0; i < points.size(); ++i)
        {
            for (int j = 0; j < N; ++j)
            {
                if (distance(current_model[j], points[i]) < sigma)
                {
                    current_inliers[j].push_back(points[i]);
                    ++inliers_nr;
                    break;
                }
            }
        }

        // 4. Save the best model so far
        if (inliers_nr > best_inliers_nr)
        {
            best_inliers_nr = inliers_nr;
            best_model = current_model;
            best_inliers = current_inliers;

            max_iterations = std::min(max_iterations, iterationNumber(0.99, 2*N, best_inliers_nr, points.size()));
            //std::cout << "New max iters: " << max_iterations << std::endl;
        }
    }

    for (int i = 0; i < N; ++i)
    {
        best_model[i] = fitLine(best_inliers[i]);
    }
    return best_model;
}

std::vector<cv::Point3f> SeqRANSAC(std::vector<cv::Point2f> points, int max_iterations, float sigma, int N)
{
    std::vector<cv::Point3f> lines;
    for (int i = 0; i < N; ++i)
    {
        cv::Point3f line = RANSAC(points, max_iterations, sigma);
        lines.push_back(line);

        std::vector<bool> isInlier(points.size(), false);
        for (int i = 0; i < points.size(); ++i)
        {
            if (distance(line, points[i]) < sigma)
            {
                isInlier[i] = true;
            }
        }

        for (int i = points.size() - 1; i >= 0; --i)
        {
            if (isInlier[i])
            {
                points.erase(std::next(points.begin(), i));
            }
        }
    }

    return lines;
}

void drawLine(cv::Point2f p, cv::Point2f dir, cv::Scalar color, cv::Mat& img)
{
    cv::Point2f p1, p2;
    p1 = p - 1000.f * dir;
    p2 = p + 1000.f * dir;
    cv::line(img, p1, p2, color, 2);
}

void drawLine(cv::Point3f l, cv::Scalar color, cv::Mat& img)
{
    cv::Point2f p1, p2;
    p1.x = 0;
    p1.y = -l.z / l.y;

    p2.x = WIDTH;
    p2.y = (-l.z - l.x * p2.x) / l.y;
    cv::line(img, p1, p2, color, 2);
}

int main()
{
    init_random();
    cv::Mat img = cv::Mat::zeros(HEIGHT, WIDTH, CV_8UC3);

    std::vector<cv::Point2f> points;
    cv::Point2f p = randomPoint();
    cv::Point2f dir = randomDir();;
    drawLine(p, dir, { 255, 0, 0 }, img);
    generateData(points, 1000, p, dir, 0.7f);
    p = randomPoint();
    dir = randomDir();
    drawLine(p, dir, { 255, 0, 0 }, img);
    generateData(points, 1000, p, dir, 0.7f);
    p = randomPoint();
    dir = randomDir();
    drawLine(p, dir, { 255, 0, 0 }, img);
    generateData(points, 1000, p, dir, 0.7f);

    for (int i = 0; i < points.size(); ++i)
    {
        cv::circle(img, points[i], 2, { 255, 255, 255 }, cv::FILLED);
    }

    std::vector<cv::Point3f> lines = SeqRANSAC(points, 2000, 5.f, 3);
    drawLine(lines[0], { 0, 255, 0 }, img);
    drawLine(lines[1], { 0, 255, 255 }, img);
    drawLine(lines[2], { 255, 0, 255 }, img);

    cv::imshow("RANSAC", img);
    int key = cv::waitKey(0);

    return 0;
}

